@@ -2,28 +2,28 @@
 import asyncio
 import pathlib
 import shutil
-import subprocess as _subprocess
+import subprocess as subprocess
 from concurrent import futures
 from functools import cached_property, partial
 from typing import (
-    AsyncIterator, Iterable, Optional,
+    AsyncIterator, Dict, Iterable, List, Optional,
     Pattern, Set, Tuple, Type, Union)
 
 import abstracts
 
-from aio.core import event, subprocess
-from aio.core.dev import debug
+from aio.core import event, subprocess as _subprocess
 from aio.core.functional import async_property, async_set, AwaitableGenerator
-from aio.core.subprocess import exceptions as subprocess_exceptions
 
 
 GREP_MIN_BATCH_SIZE = 500
 GREP_MAX_BATCH_SIZE = 2000
 
 
-# TODO: factor this into a generic subproc runner/parser
-class ADirectoryGrepper(metaclass=abstracts.Abstraction):
-    """Blocking directory grepper.
+@abstracts.implementer(_subprocess.ISubprocessHandler)
+class ADirectoryFileFinder(
+        _subprocess.ASubprocessHandler,
+        metaclass=abstracts.Abstraction):
+    """Blocking directory finder.
 
     Run inside a subproc, so *must* be picklable.
     """
@@ -42,29 +42,26 @@ class ADirectoryGrepper(metaclass=abstracts.Abstraction):
             path: str,
             path_matcher: Optional[Pattern[str]] = None,
             exclude_matcher:  Optional[Pattern[str]] = None) -> None:
-        self.path = path
+        _subprocess.ASubprocessHandler.__init__(self, path)
         self.path_matcher = path_matcher
         self.exclude_matcher = exclude_matcher
 
-    @debug.logging(
-        log=__name__,
-        show_cpu=True)
-    def __call__(self, *args: str) -> Set[str]:
-        """Call `grep` command with args."""
-        # TODO: Add error handling!!!
+    def handle(self, response: subprocess.CompletedProcess) -> Set[str]:
         return set(
             path
             for path
-            in _subprocess.run(
-                args,
-                cwd=self.path,
-                capture_output=True,
-                encoding="utf-8").stdout.split("\n")
+            in response.stdout.split("\n")
             if self.include_path(
                 path,
                 self.path_matcher,
                 self.exclude_matcher))
 
+    def handle_error(
+            self,
+            response: subprocess.CompletedProcess) -> Dict[str, List[str]]:
+        # TODO: Handle errors in directory classes
+        return super().handle_error(response)
+
 
 @abstracts.implementer(event.IExecutive)
 class ADirectory(event.AExecutive, metaclass=abstracts.Abstraction):
@@ -107,16 +104,23 @@ class ADirectory(event.AExecutive, metaclass=abstracts.Abstraction):
     def absolute_path(self) -> str:
         return str(self.path.resolve())
 
-    @property  # type:ignore
-    @abstracts.interfacemethod
-    def directory_grepper_class(self) -> Type[ADirectoryGrepper]:
-        raise NotImplementedError
-
     @async_property(cache=True)
     async def files(self) -> Set[str]:
         """Set of relative file paths associated with this directory."""
         return await self.get_files()
 
+    @cached_property
+    def finder(self) -> ADirectoryFileFinder:
+        return self.finder_class(
+            str(self.path),
+            path_matcher=self.path_matcher,
+            exclude_matcher=self.exclude_matcher)
+
+    @property  # type:ignore
+    @abstracts.interfacemethod
+    def finder_class(self) -> Type[ADirectoryFileFinder]:
+        raise NotImplementedError
+
     @cached_property
     def grep_args(self) -> Tuple[str, ...]:
         """Default args to pass when calling `grep`."""
@@ -133,7 +137,7 @@ class ADirectory(event.AExecutive, metaclass=abstracts.Abstraction):
         grep_command = shutil.which("grep")
         if grep_command:
             return grep_command, "-r"
-        raise subprocess_exceptions.OSCommandError(
+        raise _subprocess.exceptions.OSCommandError(
             "Unable to find `grep` command")
 
     @property
@@ -157,26 +161,11 @@ class ADirectory(event.AExecutive, metaclass=abstracts.Abstraction):
     def grep_min_batch_size(self) -> int:
         return GREP_MIN_BATCH_SIZE
 
-    @cached_property
-    def grepper(self) -> ADirectoryGrepper:
-        return self.directory_grepper_class(
-            str(self.path),
-            path_matcher=self.path_matcher,
-            exclude_matcher=self.exclude_matcher)
-
     @cached_property
     def path(self) -> pathlib.Path:
         """Path to this directory."""
         return pathlib.Path(self._path)
 
-    @cached_property
-    def shell(self) -> subprocess.AAsyncShell:
-        """Shell that uses the directory path as `cwd`."""
-        return subprocess.AsyncShell(
-            cwd=self.path,
-            loop=self.loop,
-            pool=self.pool)
-
     async def get_files(self) -> Set[str]:
         return await self.grep(["-l"], "")
 
@@ -208,7 +197,7 @@ class ADirectory(event.AExecutive, metaclass=abstracts.Abstraction):
             grep_args: Tuple[str, ...],
             paths: Iterable[str]) -> AwaitableGenerator:
         return self.execute_in_batches(
-            partial(self.grepper, *grep_args),
+            partial(self.finder, *grep_args),
             *paths,
             min_batch_size=self.grep_min_batch_size,
             max_batch_size=self.grep_max_batch_size)
@@ -234,6 +223,20 @@ class AGitDirectory(ADirectory):
     Uses the `git` index cache for faster grepping.
     """
 
+    @classmethod
+    def find_git_files_changed_since(
+            cls,
+            finder: ADirectoryFileFinder,
+            git_command: str,
+            since: str) -> Set[str]:
+        return finder(
+            git_command,
+            "diff",
+            "--name-only",
+            since,
+            "--diff-filter=ACMR",
+            "--ignore-submodules=all")
+
     def __init__(self, *args, **kwargs) -> None:
         self.changed = kwargs.pop("changed", None)
         super().__init__(*args, **kwargs)
@@ -242,45 +245,17 @@ class AGitDirectory(ADirectory):
     async def changed_files(self) -> Set[str]:
         """Files that have changed since `self.changed`, which can be the name
         of a git object (branch etc) or a commit hash."""
-        # TODO: move filtering into a subshell
-        return set(
-            path
-            for path
-            in await self.shell(
-                [self.git_command,
-                 "diff",
-                 "--name-only",
-                 self.changed,
-                 "--diff-filter=ACMR",
-                 "--ignore-submodules=all"])
-            if (path
-                and self.directory_grepper_class.include_path(
-                    path,
-                    self.path_matcher,
-                    self.exclude_matcher)))
+        return await self.execute(
+            self.find_git_files_changed_since,
+            self.finder,
+            self.git_command,
+            self.changed)
 
     @async_property(cache=True)
     async def files(self) -> Set[str]:
-        return (
-            await self.get_files()
-            if not self.changed
-            else await self.get_changed_files())
-
-    @property
-    def git_command(self) -> str:
-        """Path to the `git` command."""
-        git_command = shutil.which("git")
-        if git_command:
-            return git_command
-        raise subprocess_exceptions.OSCommandError(
-            "Unable to find the `git` command")
-
-    @property
-    def grep_command_args(self) -> Tuple[str, ...]:
-        return self.git_command, "grep", "--cached"
-
-    async def get_changed_files(self) -> Set[str]:
         """Files that have changed."""
+        if not self.changed:
+            return await self.get_files()
         # If the super `files` are going to be filtered fetch them first, and
         # only check `changed_files` if there are any matching. Otherwise do
         # the reverse - get the `changed_files` and bail if there are none.
@@ -296,3 +271,16 @@ class AGitDirectory(ADirectory):
              if files is None
              else files)
             & await self.changed_files)
+
+    @property
+    def git_command(self) -> str:
+        """Path to the `git` command."""
+        git_command = shutil.which("git")
+        if git_command:
+            return git_command
+        raise _subprocess.exceptions.OSCommandError(
+            "Unable to find the `git` command")
+
+    @property
+    def grep_command_args(self) -> Tuple[str, ...]:
+        return self.git_command, "grep", "--cached"
